from .tclab import TCLab, TCLabModel
from .historian import Historian, Plotter
from .labtime import labtime, clock


class Experiment:
    def __init__(self, connected=True, plot=True,
                 twindow=200, time=500,
                 speedup=1):
        self.connected = connected
        self.plot = plot
        self.twindow = twindow
        self.time = time

        self.lab = None
        self.historian = None
        self.plotter = None

        labtime.set_rate(speedup)

    def __enter__(self):
        if self.connected:
            self.lab = TCLab()
        else:
            self.lab = TCLabModel()
        self.historian = Historian(self.lab.sources)
        if self.plot:
            self.plotter = Plotter(self.historian)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.lab.close()
        self.historian.close()

    def clock(self):
        for t in clock(self.time):
            if self.plot:
                self.plotter.update(t)
            else:
                self.historian.update(t)
            yield t


def runexperiment(function, connected=True, plot=True,
                  twindow=200, time=500,
                  speedup=1):
    with Experiment(connected, plot, twindow, time, speedup) as experiment:
        for t in experiment.clock():
            function(t, experiment.lab)
    return experiment



from .tclab import TCLab, TCLabModel
from .historian import Historian, Plotter
from .experiment import Experiment, runexperiment
from .labtime import clock, labtime, setnow


def setup(connected=True, speedup=1):
    """Set up a lab session with simple switching between real and model lab

    The idea of this function is that you will do

    >>> lab = setup(connected=True)

    to obtain a TCLab class reference. If `connected=False` then you will
    receive a TCLabModel class reference. This allows you to switch between
    the model and the real lab in your code easily.

    The speedup option can only be used when `connected=False` and is the
    ratio by which the lab clock will be sped up relative to real time
    during the simulation.

    For example

    >>> lab = setup(connected=False, speedup=2)

    will run the lab clock at twice real time (which means that the whole
    simulation will take half the time it would if connected to a real device).
    """

    if connected:
        lab = TCLab
        if speedup != 1:
            raise ValueError('The real lab must run in real time')
    else:
        lab = TCLabModel
        if speedup < 0:
            raise ValueError('speedup must be positive. '
                             'You passed speedup={}'.format(speedup))

    labtime.set_rate(speedup)
    return lab



import time as time


class Labtime():
    __realtime = time.time()
    __labtime = 0
    __rate = 1
    __running = True

    @property
    def running(self):
        """Returns variable indicating whether labtime is running."""
        return self.__class__.__running

    def time(self):
        """Return current labtime."""
        if self.running:
            elapsed = time.time() - self.__class__.__realtime
            return self.__class__.__labtime + self.__class__.__rate * elapsed
        else:
            return self.__class__.__labtime

    def set_rate(self, rate=1):
        """Set the rate of labtime relative to real time."""
        if rate <= 0:
            raise ValueError("Labtime rates must be positive.")
        self.__class__.__labtime = self.time()
        self.__class__.__realtime = time.time()
        self.__class__.__rate = rate

    def get_rate(self):
        """Return the rate of labtime relative to real time."""
        return self.__class__.__rate

    def sleep(self, delay):
        """Sleep in labtime for a period delay."""
        if self.__class__.__running:
            time.sleep(delay / self.__class__.__rate)
        else:
            raise RuntimeWarning("sleep is not valid when labtime is stopped.")

    def stop(self):
        """Stop labtime."""
        self.__class__.__labtime = self.time()
        self.__class__.__realtime = time.time()
        self.__class__.__running = False

    def start(self):
        """Restart labtime."""
        self.__class__.__realtime = time.time()
        self.__class__.__running = True

    def reset(self, val=0):
        """Reset labtime to a specified value."""
        self.__class__.__labtime = val
        self.__class__.__realtime = time.time()


labtime = Labtime()


# for backwards compatability
def setnow(tnow=0):
    labtime.reset(tnow)


def clock(period, step=1, tol=0.5):
    """Generator providing time values in sync with real time clock.

    Args:
        tperiod (float): Time interval for clock operation in seconds.
        tstep (float): Time step.
        tol (float): Maximum permissible deviation from real time.

    Yields:
        float: The next time step rounded to nearest 10th of a second.
    """
    now = 0
    start = labtime.time()

    while now <= period - step + tol:
        yield round(now, 1)
        labtime.sleep(step - (labtime.time() - start) % step)
        now = labtime.time() - start

    yield round(now, 1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import print_function
import time
import os
import random
import serial
from serial.tools import list_ports
from .labtime import labtime


sep = ' '   # command/value separator in TCLab firmware

arduinos = [('USB VID:PID=16D0:0613', 'Arduino Uno'),
            ('USB VID:PID=1A86:7523', 'NHduino'),
            ('USB VID:PID=2341:8036', 'Arduino Leonardo'),
            ('USB VID:PID=2A03', 'Arduino.org device'),
            ('USB VID:PID', 'unknown device'),
            ]


def clip(val, lower=0, upper=100):
    """Limit value to be between lower and upper limits"""
    return max(lower, min(val, upper))


def command(name, argument, lower=0, upper=100):
    """Construct command to TCLab-sketch."""
    return name + sep + str(clip(argument, lower, upper))


class TCLab(object):
    def __init__(self, port='', debug=False):
        self.debug = debug
        if os.name == 'nt':
            port_list = list_ports.comports()
        else:
            port_list = list_ports.grep(port)
        for comport in port_list:
            for key, val in arduinos:
                if comport[2].startswith(key):
                    self.arduino = val
                    port = comport[0]
                    break
            else:
                continue  # key not found in arduinos
            break  # key was found in arduinos
        else:
            print('--- Serial Ports ---')
            for comport in list(list_ports.comports()):
                print(" ".join(comport))
            raise RuntimeError('No Arduino device found.')
        try:
            try:
                baud = 115200
                self.sp = serial.Serial(port=port, baudrate=baud, timeout=2)
                time.sleep(2)
                self.Q1(0)
            except:
                self.sp.close()
                baud = 9600
                self.sp = serial.Serial(port=port, baudrate=baud, timeout=2)
                time.sleep(2)
                self.Q1(0)  # fails if not connected
                print('New Arduino TCLab firmware available at:')
                print(' https://github.com/jckantor/TCLab-sketch')
        except:
            raise RuntimeError('Failed to Connect.')
        self.sp.readline().decode('UTF-8')
        self.sp.write(('VER' + '\r\n').encode())
        self.version = self.sp.readline()
        self.version = self.version.decode('UTF-8').replace('\r\n', '')
        if self.sp.isOpen():
            print(self.arduino, 'connected on port', port, 'at', baud, 'baud.')
            print(self.version + '.')
        labtime.set_rate(1)
        self._P1 = 200.0
        self._P2 = 100.0
        self.Q1(0)
        self.sources = [('T1', self.scan),
                        ('T2', None),
                        ('Q1', None),
                        ('Q2', None),
                        ]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return

    def close(self):
        """Shut down TCLab device and close serial connection."""
        self.Q1(0)
        self.Q2(0)
        self.send_and_receive('X')
        self.sp.close()
        print('TCLab disconnected successfully.')
        return

    def send(self, msg):
        """Send a string message to the TCLab firmware."""
        self.sp.write((msg + '\r\n').encode())
        if self.debug:
            print('Sent: "' + msg + '"')
        self.sp.flush()

    def receive(self):
        """Return a string message received from the TCLab firmware."""
        msg = self.sp.readline().decode('UTF-8').replace('\r\n', '')
        if self.debug:
            print('Return: "' + msg + '"')
        return msg

    def send_and_receive(self, msg, convert=str):
        """Send a string message and return the response"""
        self.send(msg)
        return convert(self.receive())

    def LED(self, val=100):
        """Flash TCLab LED at a specified brightness for 10 seconds."""
        return self.send_and_receive(command('LED', val), float)

    @property
    def T1(self):
        """Return a float denoting TCLab temperature T1 in degrees C."""
        return self.send_and_receive('T1', float)

    @property
    def T2(self):
        """Return a float denoting TCLab temperature T2 in degrees C."""
        return self.send_and_receive('T2', float)

    @property
    def P1(self):
        """Return a float denoting maximum power of heater 1 in pwm."""
        return self._P1

    @P1.setter
    def P1(self, val):
        """Set maximum power of heater 1 in pwm, range 0 to 255."""
        self._P1 = self.send_and_receive(command('P1', val, 0, 255), float)

    @property
    def P2(self):
        """Return a float denoting maximum power of heater 2 in pwm."""
        return self._P2

    @P2.setter
    def P2(self, val):
        """Set maximum power of heater 2 in pwm, range 0 to 255."""
        self._P2 = self.send_and_receive(command('P2', val, 0, 255), float)

    def Q1(self, val=None):
        """Get or set TCLab heater power Q1

        val: Value of heater power, range is limited to 0-100

        return clipped value."""
        if val is None:
            msg = 'R1'
        else:
            msg = 'Q1' + sep + str(clip(val))
        return self.send_and_receive(msg, float)

    def Q2(self, val=None):
        """Get or set TCLab heater power Q2

        val: Value of heater power, range is limited to 0-100

        return clipped value."""
        if val is None:
            msg = 'R2'
        else:
            msg = 'Q2' + sep + str(clip(val))
        return self.send_and_receive(msg, float)

    def scan(self):
        self.send('SCAN')
        T1 = float(self.receive())
        T2 = float(self.receive())
        Q1 = float(self.receive())
        Q2 = float(self.receive())
        return T1, T2, Q1, Q2

    # Define properties for Q1 and Q2
    U1 = property(fget=Q1, fset=Q1, doc="Heater 1 value")
    U2 = property(fget=Q2, fset=Q2, doc="Heater 2 value")


class TCLabModel(object):
    def __init__(self, port='', debug=False):
        self.debug = debug
        print('Simulated TCLab')
        self.Ta = 21                  # ambient temperature
        self.tstart = labtime.time()  # start time
        self.tlast = self.tstart      # last update time
        self._P1 = 200.0              # max power heater 1
        self._P2 = 100.0              # max power heater 2
        self._Q1 = 0                  # initial heater 1
        self._Q2 = 0                  # initial heater 2
        self._T1 = self.Ta            # temperature thermister 1
        self._T2 = self.Ta            # temperature thermister 2
        self._H1 = self.Ta            # temperature heater 1
        self._H2 = self.Ta            # temperature heater 2
        self.maxstep = 0.2            # maximum time step for integration
        self.sources = [('T1', self.scan),
                        ('T2', None),
                        ('Q1', None),
                        ('Q2', None),
                        ]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return

    def close(self):
        """Simulate shutting down TCLab device."""
        self.Q1(0)
        self.Q2(0)
        print('TCLab Model disconnected successfully.')
        return

    def LED(self, val=100):
        """Simulate flashing TCLab LED

           val : specified brightness (default 100). """
        self.update()
        return clip(val)

    @property
    def T1(self):
        """Return a float denoting TCLab temperature T1 in degrees C."""
        self.update()
        return self.measurement(self._T1)

    @property
    def T2(self):
        """Return a float denoting TCLab temperature T2 in degrees C."""
        self.update()
        return self.measurement(self._T2)

    @property
    def P1(self):
        """Return a float denoting maximum power of heater 1 in pwm."""
        self.update()
        return self._P1

    @P1.setter
    def P1(self, val):
        """Set maximum power of heater 1 in pwm, range 0 to 255."""
        self.update()
        self._P1 = clip(val, 0, 255)

    @property
    def P2(self):
        """Return a float denoting maximum power of heater 2 in pwm."""
        self.update()
        return self._P2

    @P2.setter
    def P2(self, val):
        """Set maximum power of heater 2 in pwm, range 0 to 255."""
        self.update()
        self._P2 = clip(val, 0, 255)

    def Q1(self, val=None):
        """Get or set TCLabModel heater power Q1

        val: Value of heater power, range is limited to 0-100

        return clipped value."""
        self.update()
        if val is not None:
            self._Q1 = clip(val)
        return self._Q1

    def Q2(self, val=None):
        """Get or set TCLabModel heater power Q2

        val: Value of heater power, range is limited to 0-100

        return clipped value."""
        self.update()
        if val is not None:
            self._Q2 = clip(val)
        return self._Q2

    def scan(self):
        self.update()
        return (self.measurement(self._T1),
                self.measurement(self._T2),
                self._Q1,
                self._Q2)

    # Define properties for Q1 and Q2
    U1 = property(fget=Q1, fset=Q1, doc="Heater 1 value")
    U2 = property(fget=Q2, fset=Q2, doc="Heater 2 value")

    def quantize(self, T):
        """Quantize model temperatures to mimic Arduino A/D conversion."""
        return max(-50, min(132.2, T - T % 0.3223))

    def measurement(self, T):
        return self.quantize(T + random.normalvariate(0, 0.043))

    def update(self):
        self.tnow = labtime.time() - self.tstart
        fullsteps, remainder = divmod(self.tnow - self.tlast, self.maxstep)

        for dt in [self.maxstep]*int(fullsteps) + [remainder]:
            DeltaTaH1 = self.Ta - self._H1
            DeltaTaH2 = self.Ta - self._H2
            DeltaT12 = self._H1 - self._H2
            dH1 = self._P1 * self._Q1 / 5720 + DeltaTaH1 / 20 - DeltaT12 / 100
            dH2 = self._P2 * self._Q2 / 5720 + DeltaTaH2 / 20 + DeltaT12 / 100
            dT1 = (self._H1 - self._T1)/140
            dT2 = (self._H2 - self._T2)/140

            self._H1 += dt * dH1
            self._H2 += dt * dH2
            self._T1 += dt * dT1
            self._T2 += dt * dT2

        self.tlast = self.tnow



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import division
from collections import Iterable
import bisect
import sqlite3
from .labtime import labtime
import time


class TagDB:
    """Interface to sqlite database containing tag values"""
    def __init__(self, filename=":memory:"):
        """Create or connect to a database

        :param filename: The filename of the database.
                         By default, values are stored in memory."""
        self.db = sqlite3.connect(filename)
        self.cursor = self.db.cursor()
        creates = ["""CREATE TABLE IF NOT EXISTS tagvalues (
                           session_id REFERENCES session (id),
                           timeseconds, name, value)""",
                   """CREATE TABLE IF NOT EXISTS sessions (
                           id INTEGER PRIMARY KEY,
                           starttime)"""]
        for statement in creates:
            self.cursor.execute(statement)
        self.db.commit()
        self.session = None

    def new_session(self):
        self.cursor.execute("""INSERT INTO SESSIONS (starttime)
                               VALUES (datetime('now'))""")
        self.session = self.cursor.lastrowid
        self.db.commit()

    def get_sessions(self):
        result = self.cursor.execute("SELECT id, starttime FROM sessions")
        return list(result)

    def record(self, timeseconds, name, value):
        if self.session is None:
            self.new_session()
        self.cursor.execute("INSERT INTO tagvalues VALUES (?, ?, ?, ?)",
                            (self.session, timeseconds, name, value))
        self.db.commit()

    def get(self, name, timeseconds=None, session=None):
        if session is None:
            session = self.session
        query = """SELECT timeseconds, value FROM tagvalues
                   WHERE session_id=? AND name=?"""
        parameters = [session, name]
        if timeseconds is not None:
            query += " and timeseconds=?"
            parameters.append(timeseconds)
        query += " ORDER BY timeseconds"
        return list(self.cursor.execute(query, parameters))

    def close(self):
        self.db.close()


class Historian(object):
    """Generalised logging class"""
    def __init__(self, sources, dbfile=":memory:"):
        """
        sources: an iterable of (name, callable) tuples
            - name (str) is the name of a signal and the
            - callable is evaluated to obtain the value.

        Example:

        >>> a = 1
        >>> def getvalue():
        ...     return a
        >>> h = Historian([('a', getvalue)])
        >>> h.update(0)
        >>> h.log
        [(0, 1)]


        Sometimes, multiple values are obtained from one function call. In such
        cases, names can still be specified as before, but callable can be
        passed as None for the subsequent names which come from a previous
        callable.

        Example:

        >>> a = 1
        >>> b = 2
        >>> def getvalues():
        ...     return [a, b]
        >>> h = Historian([('a', getvalues),
        ...                ('b', None)])
        >>> h.update(0)
        [(0, 1, 2)]

        """
        self.sources = [('Time', lambda: self.tnow)] + list(sources)
        if dbfile:
            self.db = TagDB(dbfile)
            self.db.new_session()
            self.session = self.db.session
        else:
            self.db = None
            self.session = 1

        self.tstart = labtime.time()

        self.columns = [name for name, _ in self.sources]

        self.build_fields()

    def build_fields(self):
        self.fields = [[] for _ in self.columns]
        self.logdict = dict(zip(self.columns, self.fields))
        self.t = self.logdict['Time']

    def update(self, tnow=None):
        if tnow is None:
            self.tnow = labtime.time() - self.tstart
        else:
            self.tnow = tnow

        for name, valuefunction in self.sources:
            if valuefunction:
                v = valuefunction()
                if isinstance(v, Iterable):
                    values = iter(v)
                else:
                    values = iter([v])
            try:
                value = next(values)
            except StopIteration:
                raise ValueError("valuefunction did not return enough values")

            self.logdict[name].append(value)
            if self.db and name != "Time":
                self.db.record(self.tnow, name, value)

    @property
    def log(self):
        return list(zip(*[self.logdict[c] for c in self.columns]))

    def timeindex(self, t):
        return max(bisect.bisect(self.t, t) - 1, 0)

    def timeslice(self, tstart=0, tend=None, columns=None):
        start = self.timeindex(tstart)
        if tend is None:
            stop = len(self.t) + 1
        # Ensure that we always return at least one time's value
        if tend == tstart:
            stop = start + 1
        if columns is None:
            columns = self.columns
        return [self.logdict[c][start:stop] for c in columns]

    def at(self, t, columns=None):
        """ Return the values of columns at or just before a certain time"""
        return [c[0] for c in self.timeslice(t, t, columns)]

    def after(self, t, columns=None):
        """ Return the values of columns after or just before a certain time"""
        return self.timeslice(t, columns=columns)

    def new_session(self):
        if self.db is None:
            raise NotImplemented("Sessions not supported without dbfile")
        self.db.new_session()
        self.session = self.db.session
        self.tstart = labtime.time()
        self.build_fields()

    def get_sessions(self):
        if self.db is None:
            raise NotImplemented("Sessions not supported without dbfile")
        return self.db.get_sessions()

    def load_session(self, session):
        if self.db is None:
            raise NotImplemented("Sessions not supported without dbfile")
        self.db.session = session
        self.build_fields()
        # FIXME: The way time is handled here is a bit brittle
        first = True
        for name in self.columns[1:]:
            for t, value in self.db.get(name):
                self.logdict[name].append(value)
                if first:
                    self.t.append(t)
            first = False

    def close(self):
        if self.db:
            self.db.close()


class Plotter:
    def __init__(self, historian, twindow=120, layout=None):
        """Generalised graphical output of a Historian

        :param historian: An instance of the Historian class
        :param twindow: Amount of time to show in the plot
        :param layout: A tuple of tuples indicating how the fields should be
                plotted. For example (("T1", "T2"), ("Q1", "Q2")) indicates
                that there will be two subplots with the two indicated fields
                plotted on each.
        """
        import matplotlib.pyplot as plt
        from matplotlib import get_backend
        self.backend = get_backend()
        self.historian = historian
        self.twindow = twindow
        self.last_plot_update = 0

        if layout is None:
            layout = tuple((field,) for field in historian.columns[1:])
        self.layout = layout

        line_options = {'where': 'post', 'lw': 2, 'alpha': 0.8}
        self.lines = {}
        self.fig, self.axes = plt.subplots(len(layout), 1, figsize=(8, 6),
                                           dpi = 80,
                                           sharex=True,
                                           gridspec_kw={'hspace': 0})
        values = {c: 0 for c in historian.columns}
        plt.setp([a.get_xticklabels() for a in self.axes[:-1]], visible=False)
        for axis, fields in zip(self.axes, self.layout):
            for field in fields:
                y = values[field]
                self.lines[field] = axis.step(0, y, label=field,
                                              **line_options)[0]
            axis.set_xlim(0, self.twindow)
            axis.autoscale(axis='y', tight=False)
            axis.set_ylabel(', '.join(fields))
            if len(fields) > 1:
                axis.legend()
            axis.grid()

        self.axes[-1].set_xlabel('Time / Seconds')
        plt.tight_layout()
        self.fig.canvas.draw()
        self.fig.show()
        if self.backend != 'nbAgg':
            from IPython import display
            self.display = display
            self.display.clear_output(wait=True)
            self.display.display(self.fig)

    def update(self, tnow=None):
        self.historian.update(tnow)

        if time.time() - self.last_plot_update <= 0.33:
            return

        self.last_plot_update = time.time()
        tmin = max(self.historian.tnow - self.twindow, 0)
        tmax = max(self.historian.tnow, self.twindow)
        for axis in self.axes:
            axis.set_xlim(tmin, tmax)
        data = self.historian.after(tmin)
        datadict = dict(zip(self.historian.columns, data))
        t = datadict['Time']
        for axis, fields in zip(self.axes, self.layout):
            for field in fields:
                y = datadict[field]
                self.lines[field].set_data(t, y)
            axis.relim()
            axis.autoscale_view()
        self.fig.canvas.draw()
        if self.backend != 'nbAgg':
            self.display.clear_output(wait=True)
            self.display.display(self.fig)
